"""Feature schemas for the License Manager API."""
from typing import Optional

from pydantic import BaseModel

from lm_backend.api.schemas.base import BaseCreateSchema, BaseUpdateSchema
from lm_backend.api.schemas.product import ProductSchema


class FeatureCreateSchema(BaseCreateSchema):
    """
    Represents the features in a feature configuration.
    """

    name: str
    product_id: int
    config_id: int
    reserved: int


class FeatureUpdateSchema(BaseUpdateSchema):
    """
    Represents the features in a feature configuration.
    """

    name: Optional[str] = None
    product_id: Optional[int] = None
    config_id: Optional[int] = None
    reserved: Optional[int] = None
    total: Optional[int] = None
    used: Optional[int] = None


class FeatureUpdateByNameSchema(BaseUpdateSchema):
    """
    Represents the feature usage data that will be updated using the name as a filter.
    """

    name: str
    total: int
    used: int


class FeatureSchema(BaseModel):
    """
    Represents the features in a feature configuration.
    """

    id: int
    name: str
    product: ProductSchema
    config_id: int
    reserved: int
    total: int
    used: int
    booked_total: Optional[int] = None

    class Config:
        orm_mode = True



"""Booking schemas for the License Manager API."""
from typing import Optional

from pydantic import BaseModel

from lm_backend.api.schemas.base import BaseCreateSchema, BaseUpdateSchema


class BookingCreateSchema(BaseCreateSchema):
    """
    Represents the booking of a feature.
    """

    job_id: int
    feature_id: int
    quantity: int


class BookingUpdateSchema(BaseUpdateSchema):
    """
    Represents the booking of a feature.
    """

    job_id: Optional[int] = None
    feature_id: Optional[int] = None
    quantity: Optional[int] = None


class BookingSchema(BaseModel):
    """
    Represents the booking of a feature.
    """

    id: int
    job_id: int
    feature_id: int
    quantity: int

    class Config:
        orm_mode = True



"""Job schemas for the License Manager API."""
from typing import List, Optional

from pydantic import BaseModel

from lm_backend.api.schemas.base import BaseCreateSchema, BaseUpdateSchema
from lm_backend.api.schemas.booking import BookingSchema


class JobBookingCreateSchema(BaseCreateSchema):
    product_feature: str
    quantity: int


class JobCreateSchema(BaseCreateSchema):
    """
    Represents the jobs submitted in a cluster.
    """

    slurm_job_id: str
    cluster_client_id: str
    username: str
    lead_host: str


class JobWithBookingCreateSchema(BaseCreateSchema):
    """
    Represents the jobs submitted in a cluster with its bookings.
    """

    slurm_job_id: str
    cluster_client_id: Optional[str]
    username: str
    lead_host: str
    bookings: List[JobBookingCreateSchema] = []


class JobUpdateSchema(BaseUpdateSchema):
    """
    Represents the jobs submitted in a cluster.
    """

    slurm_job_id: Optional[str]
    cluster_client_id: Optional[str]
    username: Optional[str]
    lead_host: Optional[str]


class JobSchema(BaseModel):
    """
    Represents the jobs submitted in a cluster.
    """

    id: int
    slurm_job_id: str
    cluster_client_id: str
    username: str
    lead_host: str

    bookings: Optional[List[BookingSchema]] = None

    class Config:
        orm_mode = True



"""Booking CRUD class for SQLAlchemy models."""
from fastapi import HTTPException
from loguru import logger
from sqlalchemy import func, insert, literal, select
from sqlalchemy.ext.asyncio import AsyncSession

from lm_backend.api.cruds.generic import GenericCRUD, ModelType
from lm_backend.api.models.booking import Booking
from lm_backend.api.models.feature import Feature
from lm_backend.api.schemas.booking import BookingCreateSchema


class BookingCRUD(GenericCRUD):
    """Booking CRUD module to overload create method, preventing the overbooking issue."""

    async def create(self, db_session: AsyncSession, obj=BookingCreateSchema) -> ModelType:
        """
        Create a new booking.

        Use a somewhat complex logic to check if a booking can be made and insert it in one query. In this
        case, implement the `INSERT columns FROM select_query WHERE EXISTS (SELECT subquery)` SQL paradigm
        using SQLAlchemy query API. A  booking will only be created if there are enough licenses available
        to make the booking.

        To determine if a booking can be made, we check if the amount of licenses already booked, the licenses
        in use, the licenses reserved and the amount requested, are smaller or equal the license total.
        """

        # CTE used to provide the request values to the insert/select query.
        vars_cte = select(
            literal(obj.feature_id).label("feature_id"),
            literal(obj.job_id).label("job_id"),
            literal(obj.quantity).label("quantity"),
        ).cte("vars")

        # Subquery to determine if there are enough licenses to be booked
        exists_subquery = (
            select(Feature.id)
            .select_from(Feature)
            .join(Booking, Feature.id == Booking.feature_id, isouter=True)
            .where(Feature.id == obj.feature_id)
            .group_by(Feature.id)
            .having(
                func.sum(func.coalesce(Booking.quantity, 0))
                + func.max(Feature.used)
                + func.max(Feature.reserved)
                + obj.quantity
                <= func.max(Feature.total)
            )
        ).exists()

        # Composite query using the CTE and subquery to atomically check-then-set the new booking
        insert_query = (
            insert(Booking)
            .from_select(
                ["job_id", "feature_id", "quantity"],
                select(
                    vars_cte.c.job_id,
                    vars_cte.c.feature_id,
                    vars_cte.c.quantity,
                )
                .select_from(vars_cte)
                .where(exists_subquery),
            )
            .returning(Booking)
        )

        try:
            result = await db_session.execute(insert_query)
            db_obj = result.scalars().one_or_none()
        except Exception as e:
            logger.error(e)
            raise HTTPException(status_code=400, detail="Object could not be read.")

        if db_obj is None:
            raise HTTPException(status_code=409, detail="Not enough licenses available.")

        return db_obj



"""Database model for Configurations."""
from sqlalchemy import Column, Integer, String
from sqlalchemy.orm import relationship
from sqlalchemy.sql.schema import CheckConstraint

from lm_backend.database import Base


class Configuration(Base):
    """
    Represents the feature configurations.
    """

    __tablename__ = "configs"
    id = Column(Integer, primary_key=True)
    name = Column(String, nullable=False)
    cluster_client_id = Column(String, nullable=False)
    grace_time = Column(Integer, CheckConstraint("grace_time>=0"), nullable=False)
    type = Column(String, nullable=False)

    license_servers = relationship(
        "LicenseServer", back_populates="configurations", lazy="selectin", cascade="all, delete-orphan"
    )
    features = relationship(
        "Feature", back_populates="configurations", lazy="selectin", cascade="all, delete-orphan"
    )

    searchable_fields = [name]
    sortable_fields = [name, cluster_client_id, type]

    def __repr__(self):
        return (
            f"Config(id={self.id}, "
            f"name={self.name}, "
            f"cluster_client_id={self.cluster_client_id}, "
            f"grace_time={self.grace_time}), "
            f"type={self.type}), "
        )



"""Database model for License Servers."""
from sqlalchemy import Column, Integer, String
from sqlalchemy.orm import relationship
from sqlalchemy.sql.schema import CheckConstraint, ForeignKey

from lm_backend.database import Base


class LicenseServer(Base):
    """
    Represents the license servers in a feature configuration.
    """

    __tablename__ = "license_servers"
    id = Column(Integer, primary_key=True)
    config_id = Column(Integer, ForeignKey("configs.id"), nullable=False)
    host = Column(String, nullable=False)
    port = Column(Integer, CheckConstraint("port>0"), nullable=False)

    configurations = relationship("Configuration", back_populates="license_servers", lazy="selectin")

    searchable_fields = [host]
    sortable_fields = [config_id, host]

    def __repr__(self):
        return f"LicenseServer(id={self.id}, host={self.host}, port={self.port})"



from typing import List

from sqlalchemy.ext.asyncio import AsyncSession

from lm_backend.api.cruds.generic import GenericCRUD
from lm_backend.api.models.configuration import Configuration
from lm_backend.api.models.feature import Feature
from lm_backend.api.schemas.configuration import ConfigurationCreateSchema, ConfigurationUpdateSchema
from lm_backend.api.schemas.feature import FeatureCreateSchema, FeatureUpdateSchema

crud_feature = GenericCRUD(Feature, FeatureCreateSchema, FeatureUpdateSchema)
crud_configuration = GenericCRUD(Configuration, ConfigurationCreateSchema, ConfigurationUpdateSchema)


async def find_feature_id_by_name_and_client_id(db_session: AsyncSession, feature_name: str, client_id: str):
    """
    Find the feature id by name and client id.
    """
    features: List[Feature] = await crud_feature.filter(
        db_session=db_session, filter_expressions=[Feature.name == feature_name]
    )

    for feature in features:
        configuration = await crud_configuration.read(db_session=db_session, id=feature.config_id)

        if configuration and configuration.cluster_client_id == client_id:
            return feature.id



"""
Instantiates armada-security resources for auth on api endpoints using project settings.
Also provides a factory function for TokenSecurity to reduce boilerplate.
"""

import typing

from armasec import Armasec, TokenPayload
from armasec.schemas import DomainConfig
from armasec.token_security import PermissionMode
from fastapi import Depends
from loguru import logger
from pydantic import EmailStr, root_validator

from lm_backend.config import settings


def get_domain_configs() -> typing.List[DomainConfig]:
    domain_configs = [
        DomainConfig(
            domain=settings.ARMASEC_DOMAIN,
            audience=settings.ARMASEC_AUDIENCE,
            debug_logger=logger.debug if settings.ARMASEC_DEBUG else None,
        )
    ]
    if all(
        [
            settings.ARMASEC_ADMIN_DOMAIN,
            settings.ARMASEC_ADMIN_AUDIENCE,
            settings.ARMASEC_ADMIN_MATCH_KEY,
            settings.ARMASEC_ADMIN_MATCH_VALUE,
        ]
    ):
        domain_configs.append(
            DomainConfig(
                domain=settings.ARMASEC_ADMIN_DOMAIN,
                audience=settings.ARMASEC_ADMIN_AUDIENCE,
                match_keys={settings.ARMASEC_ADMIN_MATCH_KEY: settings.ARMASEC_ADMIN_MATCH_VALUE},
                debug_logger=logger.debug if settings.ARMASEC_DEBUG else None,
            )
        )
    return domain_configs


guard = Armasec(domain_configs=get_domain_configs())


class IdentityPayload(TokenPayload):
    """
    Provide an extension of TokenPayload that includes the user's identity.
    """

    email: typing.Optional[EmailStr] = None
    organization_id: typing.Optional[str] = None

    @root_validator(pre=True)
    def extract_organization(cls, values):
        """
        Extracts the organization_id from the organization payload.

        The payload is expected to look like:
        {
            ...,
            "organization": {
                "adf99e01-5cd5-41ac-a1af-191381ad7780": {
                    ...
                }
            }
        }
        """
        organization_payload = values.pop("organization", None)
        if organization_payload is None:
            return values
        if not isinstance(organization_payload, dict) and not isinstance(organization_payload, str):
            raise ValueError(f"Invalid organization payload: {organization_payload}")
        elif len(organization_payload) != 1 and isinstance(organization_payload, dict):
            raise ValueError(
                f"Organization payload did not include exactly one value: {organization_payload}"
            )
        return {
            **values,
            "organization_id": next(iter(organization_payload))
            if isinstance(organization_payload, dict)
            else organization_payload,
        }


def lockdown_with_identity(*scopes: str, permission_mode: PermissionMode = PermissionMode.ALL):
    """
    Provide a wrapper to be used with dependency injection to extract identity on a secured route.
    """

    def dependency(
        token_payload: TokenPayload = Depends(guard.lockdown(*scopes, permission_mode=permission_mode))
    ) -> IdentityPayload:
        """
        Provide an injectable function to lockdown a route and extract the identity payload.
        """
        return IdentityPayload.parse_obj(token_payload)

    return dependency



"""License Server schemas for the License Manager API."""
from typing import Optional

from pydantic import BaseModel, PositiveInt

from lm_backend.api.schemas.base import BaseCreateSchema, BaseUpdateSchema


class LicenseServerCreateSchema(BaseCreateSchema):
    """
    License server to be created in the database.
    """

    config_id: int
    host: str
    port: PositiveInt


class LicenseServerUpdateSchema(BaseUpdateSchema):
    """
    License server to be updated in the database.
    """

    config_id: Optional[int] = None
    host: Optional[str] = None
    port: Optional[PositiveInt] = None


class LicenseServerSchema(BaseModel):
    """
    License server response from the database.
    """

    id: int
    config_id: int
    host: str
    port: PositiveInt

    class Config:
        orm_mode = True



"""Base schemas for the License Manager API."""


from pydantic import BaseModel


class BaseCreateSchema(BaseModel):
    """
    Base class for creating objects in the database.
    """

    pass


class BaseUpdateSchema(BaseModel):
    """
    Base class for updating objects in the database.
    """

    pass



from fastapi import APIRouter

from lm_backend.api.routes.bookings import router as router_bookings
from lm_backend.api.routes.configurations import router as router_configurations
from lm_backend.api.routes.features import router as router_features
from lm_backend.api.routes.jobs import router as router_jobs
from lm_backend.api.routes.license_servers import router as router_license_servers
from lm_backend.api.routes.products import router as router_products

api = APIRouter()
api.include_router(router_configurations, prefix="/configurations", tags=["Configuration"])
api.include_router(router_license_servers, prefix="/license_servers", tags=["License Server"])
api.include_router(router_products, prefix="/products", tags=["Product"])
api.include_router(router_features, prefix="/features", tags=["Feature"])
api.include_router(router_jobs, prefix="/jobs", tags=["Job"])
api.include_router(router_bookings, prefix="/bookings", tags=["Booking"])



from enum import Enum


class Permissions(str, Enum):
    """
    Describe the permissions that may be used for protecting LM Backend routes.
    """

    LICENSE_SERVER_VIEW = "license-manager:license-server:view"
    LICENSE_SERVER_EDIT = "license-manager:license-server:edit"

    CLUSTER_VIEW = "license-manager:cluster:view"
    CLUSTER_EDIT = "license-manager:cluster:edit"

    CONFIG_VIEW = "license-manager:config:view"
    CONFIG_EDIT = "license-manager:config:edit"

    PRODUCT_VIEW = "license-manager:product:view"
    PRODUCT_EDIT = "license-manager:product:edit"

    FEATURE_VIEW = "license-manager:feature:view"
    FEATURE_EDIT = "license-manager:feature:edit"

    JOB_VIEW = "license-manager:job:view"
    JOB_EDIT = "license-manager:job:edit"

    BOOKING_VIEW = "license-manager:booking:view"
    BOOKING_EDIT = "license-manager:booking:edit"



from typing import List, Optional

from fastapi import APIRouter, Depends, Query, status

from lm_backend.api.cruds.generic import GenericCRUD
from lm_backend.api.models.product import Product
from lm_backend.api.schemas.product import ProductCreateSchema, ProductSchema, ProductUpdateSchema
from lm_backend.database import SecureSession, secure_session
from lm_backend.permissions import Permissions

router = APIRouter()


crud_product = GenericCRUD(Product, ProductCreateSchema, ProductUpdateSchema)


@router.post(
    "",
    response_model=ProductSchema,
    status_code=status.HTTP_201_CREATED,
)
async def create_product(
    product: ProductCreateSchema,
    secure_session: SecureSession = Depends(secure_session(Permissions.PRODUCT_EDIT)),
):
    """Create a new product."""
    return await crud_product.create(db_session=secure_session.session, obj=product)


@router.get(
    "",
    response_model=List[ProductSchema],
    status_code=status.HTTP_200_OK,
)
async def read_all_products(
    search: Optional[str] = Query(None),
    sort_field: Optional[str] = Query(None),
    sort_ascending: bool = Query(True),
    secure_session: SecureSession = Depends(secure_session(Permissions.PRODUCT_VIEW)),
):
    """Return all products with associated features."""
    return await crud_product.read_all(
        db_session=secure_session.session, search=search, sort_field=sort_field, sort_ascending=sort_ascending
    )


@router.get(
    "/{product_id}",
    response_model=ProductSchema,
    status_code=status.HTTP_200_OK,
)
async def read_product(
    product_id: int,
    secure_session: SecureSession = Depends(secure_session(Permissions.PRODUCT_VIEW)),
):
    """Return a product with associated features with the given id."""
    return await crud_product.read(db_session=secure_session.session, id=product_id)


@router.put(
    "/{product_id}",
    response_model=ProductSchema,
    status_code=status.HTTP_200_OK,
)
async def update_product(
    product_id: int,
    product_update: ProductUpdateSchema,
    secure_session: SecureSession = Depends(secure_session(Permissions.PRODUCT_EDIT)),
):
    """Update a product in the database."""
    return await crud_product.update(
        db_session=secure_session.session,
        id=product_id,
        obj=product_update,
    )


@router.delete(
    "/{product_id}",
    status_code=status.HTTP_200_OK,
)
async def delete_product(
    product_id: int,
    secure_session: SecureSession = Depends(secure_session(Permissions.PRODUCT_EDIT)),
):
    """Delete a product from the database and associated features."""
    return await crud_product.delete(db_session=secure_session.session, id=product_id)



"""Configuration schemas for the License Manager API."""
from typing import List, Optional

from pydantic import BaseModel

from lm_backend.api.schemas.base import BaseCreateSchema, BaseUpdateSchema
from lm_backend.api.schemas.feature import FeatureSchema
from lm_backend.api.schemas.license_server import LicenseServerSchema
from lm_backend.constants import LicenseServerType


class ConfigurationCreateSchema(BaseCreateSchema):
    """
    Represents the configuration for a set of features.
    """

    name: str
    cluster_client_id: str
    grace_time: int
    type: LicenseServerType


class ConfigurationUpdateSchema(BaseUpdateSchema):
    name: Optional[str] = None
    cluster_client_id: Optional[str] = None
    grace_time: Optional[int] = None
    type: Optional[LicenseServerType] = None


class ConfigurationSchema(BaseModel):
    """
    Represents the configuration for a set of features.
    """

    id: int
    name: str
    cluster_client_id: str
    features: List[FeatureSchema] = []
    license_servers: List[LicenseServerSchema] = []
    grace_time: int
    type: LicenseServerType

    class Config:
        orm_mode = True



"""Database model for Features."""
from sqlalchemy import Column, Integer, SQLColumnExpression, String, func, select
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship
from sqlalchemy.sql.schema import CheckConstraint, ForeignKey

from lm_backend.api.models.booking import Booking
from lm_backend.database import Base


class Feature(Base):
    """
    Represents a feature.
    """

    __tablename__ = "features"
    id = Column(Integer, primary_key=True)
    name = Column(String, nullable=False)
    product_id = Column(Integer, ForeignKey("products.id"), nullable=False)
    config_id = Column(Integer, ForeignKey("configs.id"), nullable=False)
    total = Column(Integer, CheckConstraint("total>=0"), default=0, nullable=False)
    used = Column(Integer, CheckConstraint("used>=0"), default=0, nullable=False)
    reserved = Column(Integer, CheckConstraint("reserved>=0"), nullable=False)

    product = relationship("Product", back_populates="features", lazy="selectin")

    bookings = relationship(
        "Booking", back_populates="feature", lazy="selectin", cascade="all, delete-orphan"
    )
    configurations = relationship("Configuration", back_populates="features", lazy="selectin")

    searchable_fields = [name]
    sortable_fields = [name, product_id, config_id]

    @hybrid_property
    def booked_total(self) -> int:
        """
        Compute the sum of all bookings.
        """
        return sum(b.quantity for b in self.bookings)

    @booked_total.inplace.expression
    @classmethod
    def _booked_total_expression(cls) -> SQLColumnExpression[int]:
        """
        Provide a sql expression for computing the total bookings for each feature in a subquery.
        """
        return select(func.sum(Booking.quantity)).where(Booking.feature_id == cls.id).label("booked_total")

    def __repr__(self):
        return (
            f"Feature(id={self.id}, "
            f"name={self.name}, "
            f"product_id={self.product_id}, "
            f"config_id={self.config_id}, "
            f"total={self.total}, "
            f"used={self.used}, "
            f"reserved={self.reserved}, "
            f"booked_total={self.booked_total})"
        )



from typing import List, Optional

from fastapi import APIRouter, Depends, HTTPException, Query, status

from lm_backend.api.cruds.generic import GenericCRUD
from lm_backend.api.models.configuration import Configuration
from lm_backend.api.schemas.configuration import (
    ConfigurationCreateSchema,
    ConfigurationSchema,
    ConfigurationUpdateSchema,
)
from lm_backend.database import SecureSession, secure_session
from lm_backend.permissions import Permissions

router = APIRouter()


crud = GenericCRUD(Configuration, ConfigurationCreateSchema, ConfigurationUpdateSchema)


@router.post(
    "",
    response_model=ConfigurationSchema,
    status_code=status.HTTP_201_CREATED,
)
async def create_configuration(
    configuration: ConfigurationCreateSchema,
    secure_session: SecureSession = Depends(secure_session(Permissions.CONFIG_EDIT)),
):
    """Create a new configuration."""
    return await crud.create(db_session=secure_session.session, obj=configuration)


@router.get(
    "",
    response_model=List[ConfigurationSchema],
    status_code=status.HTTP_200_OK,
)
async def read_all_configurations(
    search: Optional[str] = Query(None),
    sort_field: Optional[str] = Query(None),
    sort_ascending: bool = Query(True),
    secure_session: SecureSession = Depends(secure_session(Permissions.CONFIG_VIEW)),
):
    """Return all configurations with the associated license servers and features."""
    return await crud.read_all(
        db_session=secure_session.session,
        search=search,
        sort_field=sort_field,
        sort_ascending=sort_ascending,
        force_refresh=True,
    )


@router.get(
    "/by_client_id",
    response_model=List[ConfigurationSchema],
    status_code=status.HTTP_200_OK,
)
async def read_configurations_by_client_id(
    secure_session: SecureSession = Depends(secure_session(Permissions.CONFIG_VIEW)),
):
    """Return the configurations with the specified client_id."""
    client_id = secure_session.identity_payload.client_id

    if not client_id:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=("Couldn't find a valid client_id in the access token."),
        )

    return await crud.filter(
        db_session=secure_session.session, filter_expressions=[Configuration.cluster_client_id == client_id]
    )


@router.get(
    "/{configuration_id}",
    response_model=ConfigurationSchema,
    status_code=status.HTTP_200_OK,
)
async def read_configuration(
    configuration_id: int,
    secure_session: SecureSession = Depends(secure_session(Permissions.CONFIG_VIEW)),
):
    """Return a configuration with the associated license severs and features with a given id."""
    return await crud.read(db_session=secure_session.session, id=configuration_id, force_refresh=True)


@router.put(
    "/{configuration_id}",
    response_model=ConfigurationSchema,
    status_code=status.HTTP_200_OK,
)
async def update_configuration(
    configuration_id: int,
    configuration_update: ConfigurationUpdateSchema,
    secure_session: SecureSession = Depends(secure_session(Permissions.CONFIG_EDIT)),
):
    """Update a configuration in the database."""
    return await crud.update(
        db_session=secure_session.session,
        id=configuration_id,
        obj=configuration_update,
    )


@router.delete(
    "/{configuration_id}",
    status_code=status.HTTP_200_OK,
)
async def delete_configuration(
    configuration_id: int,
    secure_session: SecureSession = Depends(secure_session(Permissions.CONFIG_EDIT)),
):
    """
    Delete a configuration from the database.

    This will also delete the features and license servers associated.
    """
    return await crud.delete(db_session=secure_session.session, id=configuration_id)



"""Database model for Bookings."""
from sqlalchemy import Column, Integer, func
from sqlalchemy.orm import relationship
from sqlalchemy.sql.schema import CheckConstraint, ForeignKey
from sqlalchemy.sql.sqltypes import DateTime

from lm_backend.database import Base


class Booking(Base):
    """
    Represents the bookings of a feature.
    """

    __tablename__ = "bookings"
    id = Column(Integer, primary_key=True)
    job_id = Column(Integer, ForeignKey("jobs.id"), nullable=False)
    feature_id = Column(Integer, ForeignKey("features.id"), nullable=False)
    quantity = Column(Integer, CheckConstraint("quantity>=0"), nullable=False)
    created_at = Column(DateTime, default=func.now())

    job = relationship("Job", back_populates="bookings", lazy="selectin")
    feature = relationship("Feature", back_populates="bookings", lazy="selectin")

    sortable_fields = [job_id, feature_id]

    def __repr__(self):
        return (
            f"Booking(id={self.id}, "
            f"job_id={self.job_id}, "
            f"feature_id={self.feature_id}, "
            f"quantity={self.quantity}, "
            f"created_at={self.created_at})"
        )



from typing import List, Optional

from fastapi import APIRouter, Depends, Query, status

from lm_backend.api.cruds.generic import GenericCRUD
from lm_backend.api.models.license_server import LicenseServer
from lm_backend.api.schemas.license_server import (
    LicenseServerCreateSchema,
    LicenseServerSchema,
    LicenseServerUpdateSchema,
)
from lm_backend.constants import LicenseServerType
from lm_backend.database import SecureSession, secure_session
from lm_backend.permissions import Permissions

router = APIRouter()


crud_license_server = GenericCRUD(LicenseServer, LicenseServerCreateSchema, LicenseServerUpdateSchema)


@router.post(
    "",
    response_model=LicenseServerSchema,
    status_code=status.HTTP_201_CREATED,
)
async def create_license_server(
    license_server: LicenseServerCreateSchema,
    secure_session: SecureSession = Depends(secure_session(Permissions.LICENSE_SERVER_EDIT)),
):
    """Create a new license server."""
    return await crud_license_server.create(db_session=secure_session.session, obj=license_server)


@router.get(
    "",
    response_model=List[LicenseServerSchema],
    status_code=status.HTTP_200_OK,
)
async def read_all_license_servers(
    search: Optional[str] = Query(None),
    sort_field: Optional[str] = Query(None),
    sort_ascending: bool = Query(True),
    secure_session: SecureSession = Depends(secure_session(Permissions.LICENSE_SERVER_VIEW)),
):
    """Return all license servers."""
    return await crud_license_server.read_all(
        db_session=secure_session.session, search=search, sort_field=sort_field, sort_ascending=sort_ascending
    )


@router.get(
    "/{license_server_id}",
    response_model=LicenseServerSchema,
    status_code=status.HTTP_200_OK,
)
async def read_license_server(
    license_server_id: int,
    secure_session: SecureSession = Depends(secure_session(Permissions.LICENSE_SERVER_VIEW)),
):
    """Return a license server with the given id."""
    return await crud_license_server.read(db_session=secure_session.session, id=license_server_id)


@router.put(
    "/{license_server_id}",
    response_model=LicenseServerSchema,
    status_code=status.HTTP_200_OK,
)
async def update_license_server(
    license_server_id: int,
    license_server_update: LicenseServerUpdateSchema,
    secure_session: SecureSession = Depends(secure_session(Permissions.LICENSE_SERVER_EDIT)),
):
    """Update a license server in the database."""
    return await crud_license_server.update(
        db_session=secure_session.session,
        id=license_server_id,
        obj=license_server_update,
    )


@router.delete(
    "/{license_server_id}",
    status_code=status.HTTP_200_OK,
)
async def delete_license_server(
    license_server_id: int,
    secure_session: SecureSession = Depends(secure_session(Permissions.LICENSE_SERVER_EDIT)),
):
    """Delete a license server from the database."""
    return await crud_license_server.delete(db_session=secure_session.session, id=license_server_id)


@router.get(
    "/types/",
    status_code=status.HTTP_200_OK,
    response_model=List[LicenseServerType],
)
async def get_license_server_types():
    """Return a list of the available license server types."""
    return LicenseServerType



""" License-manager backend, command line entrypoint

Run with e.g. `uvicorn lm_backend.main:app` OR
set `licensemanager2.backend.main.handler` as the ASGI handler
"""
import logging
import sys
from contextlib import asynccontextmanager
from typing import cast

import sentry_sdk
from fastapi import FastAPI, Response, status
from fastapi.middleware.cors import CORSMiddleware
from loguru import logger
from sentry_sdk.integrations.asgi import SentryAsgiMiddleware

from lm_backend import __version__
from lm_backend.api import api
from lm_backend.config import settings
from lm_backend.database import engine_factory

subapp = FastAPI(
    title="License Manager API",
    version=__version__,
    contact={
        "name": "Omnivector Solutions",
        "url": "https://www.omnivector.solutions/",
        "email": "info@omnivector.solutions",
    },
    license_info={
        "name": "MIT License",
        "url": "https://github.com/omnivector-solutions/license-manager/blob/main/LICENSE",
    },
    root_path=settings.ASGI_ROOT_PATH,
)

subapp.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

subapp.include_router(api)

if settings.SENTRY_DSN:
    sentry_sdk.init(
        dsn=settings.SENTRY_DSN,
        sample_rate=cast(float, settings.SENTRY_SAMPLE_RATE),  # The cast silences mypy
        environment=settings.DEPLOY_ENV,
    )
    subapp.add_middleware(SentryAsgiMiddleware)


@subapp.get(
    "/health",
    status_code=status.HTTP_204_NO_CONTENT,
    responses={204: {"description": "API is healthy"}},
)
async def health_check():
    return Response(status_code=status.HTTP_204_NO_CONTENT)


@subapp.get(
    "/version",
    status_code=status.HTTP_200_OK,
    responses={200: {"description": "API version"}},
)
async def get_version():
    return __version__


app = FastAPI()
app.mount("/lm", subapp)


@asynccontextmanager
async def lifespan(_: FastAPI):
    """
    Provide a lifespan context for the app.

    Will set up logging and cleanup database engines when the app is shut down.

    This is the preferred method of handling lifespan events in FastAPI.
    For more details, see: https://fastapi.tiangolo.com/advanced/events/
    """
    logger.remove()
    logger.add(sys.stderr, level=settings.LOG_LEVEL)
    logger.info(f"Logging configured 📝 Level: {settings.LOG_LEVEL}")

    if settings.LOG_LEVEL_SQL:
        level_sql = getattr(logging, settings.LOG_LEVEL_SQL)

        engine_logger = logging.getLogger("sqlalchemy.engine")
        engine_logger.setLevel(level_sql)

        logger.info(f"Database logging configured 📝 Level: {settings.LOG_LEVEL_SQL}")

    yield

    await engine_factory.cleanup()



"""Product schemas for the License Manager API."""
from typing import Optional

from pydantic import BaseModel

from lm_backend.api.schemas.base import BaseCreateSchema, BaseUpdateSchema


class ProductCreateSchema(BaseCreateSchema):
    """
    Represents a feature's product.
    """

    name: str


class ProductUpdateSchema(BaseUpdateSchema):
    """
    Represents a feature's product.
    """

    name: Optional[str] = None


class ProductSchema(BaseModel):
    """
    Represents a feature's product.
    """

    id: int
    name: str

    class Config:
        orm_mode = True



"""Generic CRUD class for SQLAlchemy models."""
from asyncio import gather
from typing import List, Optional, Type, TypeVar, Union

from fastapi import HTTPException
from loguru import logger
from sqlalchemy import Column, ColumnElement, and_, select
from sqlalchemy.ext.asyncio import AsyncSession

from lm_backend.api.schemas.base import BaseCreateSchema, BaseUpdateSchema
from lm_backend.database import Base, search_clause, sort_clause

ModelType = TypeVar("ModelType", bound=Base)
CreateSchemaType = TypeVar("CreateSchemaType", bound=BaseCreateSchema)
UpdateSchemaType = TypeVar("UpdateSchemaType", bound=BaseUpdateSchema)


class GenericCRUD:
    """Generic CRUD module to interface with database, to be utilized by all models."""

    def __init__(
        self,
        model: Type[ModelType],
        create_schema: Type[CreateSchemaType],
        update_schema: Type[UpdateSchemaType],
    ):
        """Initializes the CRUD class with the model to be used."""
        self.model = model
        self.create_schema = create_schema
        self.update_schema = update_schema

    async def create(self, db_session: AsyncSession, obj: CreateSchemaType) -> ModelType:
        """Creates a new object in the database."""
        db_obj = self.model(**obj.dict())
        try:
            db_session.add(db_obj)
        except Exception as e:
            logger.error(e)
            raise HTTPException(status_code=400, detail=f"{self.model.__name__} could not be created.")

        # TODO: Determine if the session actually needs to be flushed here. I think it might not
        await db_session.flush()
        await db_session.refresh(db_obj)
        return db_obj

    async def filter(
        self, db_session: AsyncSession, filter_expressions: List[ColumnElement[bool]]
    ) -> List[ModelType]:
        """
        Filter objects using a filter field and filter term.
        Returns the list of objects or raise an exception if it does not exist.
        """
        try:
            query = await db_session.execute(select(self.model).filter(and_(*filter_expressions)))
            db_objs = list(query.scalars().all())
        except Exception as e:
            logger.error(e)
            raise HTTPException(status_code=400, detail=f"{self.model.__name__} could not be read.")

        if db_objs is None:
            raise HTTPException(status_code=404, detail=f"{self.model.__name__} not found.")

        return db_objs

    async def read(
        self, db_session: AsyncSession, id: Union[Column[int], int], force_refresh: bool = False
    ) -> Optional[ModelType]:
        """
        Read an object from the database with the given id.
        Returns the object or raise an exception if it does not exist.
        """
        try:
            query = await db_session.execute(select(self.model).filter(self.model.id == id))
            db_obj = query.scalars().one_or_none()
        except Exception as e:
            logger.error(e)
            raise HTTPException(status_code=400, detail=f"{self.model.__name__} could not be read.")

        if db_obj is None:
            raise HTTPException(status_code=404, detail=f"{self.model.__name__} not found.")

        if force_refresh:
            await db_session.refresh(db_obj)

        return db_obj

    async def read_all(
        self,
        db_session: AsyncSession,
        search: str = None,
        sort_field: Optional[str] = None,
        sort_ascending: bool = True,
        force_refresh: bool = False,
    ) -> List[ModelType]:
        """
        Read all objects.
        Returns a list of objects.

        Note: The ``force_refresh`` parameter may be useful to force the gathered results to lazily load
              relationships or other computed fields like column properties or hybrid attributes.
        """
        try:
            stmt = select(self.model)
            if search is not None:
                stmt = stmt.where(search_clause(search, self.model.searchable_fields))
            if sort_field is not None:
                stmt = stmt.order_by(sort_clause(sort_field, self.model.sortable_fields, sort_ascending))
            query = await db_session.scalars(stmt)
            db_objs = list(query.all())
            if force_refresh:
                await gather(*(db_session.refresh(db_obj) for db_obj in db_objs))
        except Exception as e:
            logger.error(e)
            raise HTTPException(status_code=400, detail=f"{self.model.__name__}s could not be read.")
        return db_objs

    async def update(
        self,
        db_session: AsyncSession,
        id: Union[Column[int], int],
        obj: UpdateSchemaType,
    ) -> Optional[ModelType]:
        """
        Update an object in the database.
        Returns the updated object.
        """
        try:
            query = await db_session.execute(select(self.model).filter(self.model.id == id))
            db_obj = query.scalar_one_or_none()
        except Exception as e:
            logger.error(e)
            raise HTTPException(status_code=400, detail=f"{self.model.__name__} could not be updated.")

        if db_obj is None:
            raise HTTPException(status_code=404, detail=f"{self.model.__name__} not found.")

        if all(value is None for _, value in obj):
            raise HTTPException(
                status_code=400, detail=f"Please provide a valid field to update {self.model.__name__}."
            )

        for field, value in obj:
            if hasattr(db_obj, field) and value is not None:
                setattr(db_obj, field, value)

        try:
            await db_session.flush()
            await db_session.refresh(db_obj)
        except Exception as e:
            logger.error(e)
            raise HTTPException(status_code=400, detail=f"{self.model.__name__} could not be updated.")

        return db_obj

    async def delete(self, db_session: AsyncSession, id: Union[Column[int], int]):
        """
        Delete an object from the database.
        """
        try:
            query = await db_session.execute(select(self.model).filter(self.model.id == id))
            db_obj = query.scalar_one_or_none()
        except Exception as e:
            logger.error(e)
            raise HTTPException(status_code=400, detail=f"{self.model.__name__} could not be deleted.")

        if db_obj is None:
            raise HTTPException(status_code=404, detail=f"{self.model.__name__} not found.")

        try:
            await db_session.delete(db_obj)
            await db_session.flush()
        except Exception as e:
            logger.error(e)
            raise HTTPException(status_code=400, detail=f"{self.model.__name__} could not be deleted.")

        return {"message": f"{self.model.__name__} deleted successfully."}



"""Database model for Jobs."""
from sqlalchemy import Column, Integer, String
from sqlalchemy.orm import relationship

from lm_backend.database import Base


class Job(Base):
    """
    Represents the jobs submitted in a cluster.
    """

    __tablename__ = "jobs"
    id = Column(Integer, primary_key=True)
    slurm_job_id = Column(String, nullable=False)
    cluster_client_id = Column(String, nullable=False)
    username = Column(String, nullable=False)
    lead_host = Column(String, nullable=False)

    bookings = relationship("Booking", back_populates="job", lazy="selectin", cascade="all, delete-orphan")

    searchable_fields = [slurm_job_id, username, lead_host]
    sortable_fields = [slurm_job_id, cluster_client_id, username, lead_host]

    def __repr__(self):
        return (
            f"Job(id={self.id}, "
            f"slurm_job_id={self.slurm_job_id}, "
            f"cluster_client_id={self.cluster_client_id}, "
            f"username={self.username}, "
            f"lead_host={self.lead_host})"
        )



"""Database model for Products."""
from sqlalchemy import Column, Integer, String
from sqlalchemy.orm import relationship

from lm_backend.database import Base


class Product(Base):
    """
    Represents a feature's product.
    """

    __tablename__ = "products"
    id = Column(Integer, primary_key=True)
    name = Column(String, nullable=False, unique=True)

    features = relationship(
        "Feature", back_populates="product", lazy="selectin", cascade="all, delete-orphan"
    )

    searchable_fields = [name]
    sortable_fields = [name]

    def __repr__(self):
        return f"Product(id={self.id}, name={self.name})"
